import datetime
import logging
import logging.handlers

import pytz
import requests

from . import show

logger = logging.getLogger(__name__)


class ShowClientError(Exception):
    """ShowClient related exception."""

    pass


class ShowClient:
    """Client which fetches the show informations from the LibreTime now-playing v2 endpoint.

    Every show has a name, a start and endtime and an optional URL.
    """

    __DEFAULT_SHOW_DURATION = 30  # 30 seconds

    def __init__(self, current_show_url):

        self.current_show_url = current_show_url

        self.show = show.Show()

    def get_show_info(self, force_update=False):
        """Return a Show object."""

        if force_update:
            self.update()
        else:
            self.lazy_update()

        return self.show

    def lazy_update(self):
        # only update the info if we expect that a new show has started
        if datetime.datetime.now(pytz.timezone("UTC")) > self.show.endtime:
            logger.info("Show expired, going to update show info")
            self.update()

        else:
            logger.debug("Show still running, won't update show info")

    def update(self):
        self.show = show.Show()  # Create a new show object

        # Set the show's default end time to now + 30 seconds to prevent updates
        # happening every second and hammering the web service if something
        # goes wrong later.
        self.show.set_endtime(
            datetime.datetime.now(pytz.timezone("UTC"))
            + datetime.timedelta(seconds=self.__DEFAULT_SHOW_DURATION)
        )

        try:
            # try to get the current show informations from loopy's cast web
            # service
            data = requests.get(self.current_show_url).json()

        except Exception as e:
            logger.error("%s: Unable to get current show informations" % self.__class__)

            logger.exception(e)
            # LSB 2017: ignoring missing show update
            # raise ShowClientError('Unable to get show informations: %s' % e)
            return

        if not data["shows"]["current"]:
            # ignore if no current show is playing
            return

        # get the name of the show, aka real_name
        # ex.: Stereo Freeze
        real_name = data["shows"]["current"]["name"]

        if len(real_name) == 0:
            # keep the default show information
            logger.error("%s: No show name found" % self.__class__)
            raise ShowClientError("Missing show name")

        self.show.set_name(real_name)

        showtz = pytz.timezone(data["station"]["timezone"])

        # get the show's end time in order to time the next lookup.
        # ex.: 2012-04-28 19:00:00 (missing a tzoffset and localized!)
        end_time = data["shows"]["current"]["ends"]

        if len(end_time) == 0:
            logger.error("%s: No end found" % self.__class__)
            raise ShowClientError("Missing show end time")

        endtime = showtz.localize(
            datetime.datetime.strptime(end_time, "%Y-%m-%d %H:%M:%S")
        )

        # store as UTC datetime object
        self.show.set_endtime(endtime.astimezone(pytz.timezone("UTC")))

        # get the show's start time
        # ex.: 2012-04-28 18:00:00
        start_time = data["shows"]["current"]["starts"]

        if len(start_time) == 0:
            logger.error("%s: No start found" % self.__class__)
            raise ShowClientError("Missing show start time")

        starttime = showtz.localize(
            datetime.datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S")
        )

        # store as UTC datetime object
        self.show.set_starttime(starttime.astimezone(pytz.timezone("UTC")))

        # Check if the endtime is in the past
        # This prevents stale (wrong) show informations from beeing pushed to
        # the live stream and stops hammering the service every second
        if self.show.endtime < datetime.datetime.now(pytz.timezone("UTC")):
            logger.error(
                "%s: Show endtime %s is in the past"
                % (self.__class__, self.show.endtime)
            )

            raise ShowClientError(
                "Show end time (%s) is in the past" % self.show.endtime
            )

        # get the show's URL
        # ex.: http://www.rabe.ch/sendungen/entertainment/onda-libera.html
        url = data["shows"]["current"]["url"]

        if len(url) == 0:
            logger.error("%s: No url found" % self.__class__)
        else:
            self.show.set_url(url)

        logger.info(
            'Show "%s" started and runs from %s till %s'
            % (self.show.name, starttime, endtime)
        )

        logger.info(self.show)



import datetime
import logging
import logging.handlers
import uuid

import pytz

logger = logging.getLogger(__name__)

DEFAULT_SHOW_URL = "https://www.rabe.ch"


class ShowError(Exception):
    """Show related exception."""

    pass


class Show:
    """Show object which has a start and end time and an optional URL."""

    def __init__(self):
        self.name = None

        self.url = DEFAULT_SHOW_URL

        # The show's global unique identifier
        self.uuid = str(uuid.uuid4())

        # Get current datetime object in UTC timezone
        now = datetime.datetime.now(pytz.timezone("UTC"))

        # The show's start time, initially set to now
        self.starttime = now

        # The show's end time, initially set to to now
        self.endtime = now

    def set_name(self, name):
        # The name of the show
        self.name = name

    def set_url(self, url):
        # The URL of the show
        self.url = url

    def set_starttime(self, starttime):
        if not isinstance(starttime, datetime.datetime):
            raise ShowError("starttime has to be a datatime object")

        # The show's start time as a datetime object
        self.starttime = starttime

    def set_endtime(self, endtime):
        if not isinstance(endtime, datetime.datetime):
            raise ShowError("endtime has to be a datatime object")

        # The show's end time as a datetime object
        self.endtime = endtime

    def __str__(self):
        return "Show '%s' (%s), start: '%s', end: '%s', url: %s" % (
            self.name,
            self.uuid,
            self.starttime,
            self.endtime,
            self.url,
        )



import calendar
import datetime
import logging
import logging.handlers
import urllib.error
import urllib.parse
import urllib.request
import uuid
import warnings
from abc import ABC, abstractmethod

import isodate
import lxml.builder
import lxml.etree
import pylast
import pytz

logger = logging.getLogger(__name__)


class TrackObserver(ABC):
    """Abstract base class for all TrackObservers."""

    name = "TrackObserver"

    def get_name(self):
        return self.name

    @abstractmethod
    def track_started(self, track):  # pragma: no cover
        pass

    @abstractmethod
    def track_finished(self, track):  # pragma: no cover
        pass


class ScrobblerTrackObserver(TrackObserver):  # pragma: no cover
    """Scrobble track information to last.fm and libre.fm."""

    name = "Audioscrobbler"

    lastfm_api_key = ""
    lastfm_api_secret = ""

    librefm_api_key = ""
    librefm_api_secret = ""

    def __init__(self, user, password):
        raise Exception("Unmaintained ScrobblerTrackObserver called")

        password = pylast.md5(password)

        self.lastfm = pylast.get_lastfm_network("", "", None, user, password)
        self.librefm = pylast.get_librefm_network("", "", None, user, password)

        self.scrobblers = {
            "lastfm": self.lastfm.get_scrobbler("tst", "1.0"),
            "librefm": self.librefm.get_scrobbler("tst", "1.0"),
        }

    def track_started(self, track):

        if track.has_default_artist() or track.has_default_title():
            logger.error("Default artist or title, skip now-playing scrobble")
            return False

        logger.info(
            "AS now-playing notification for track: %s - %s"
            % (track.artist, track.title)
        )

        # scrobble to all networks
        #
        # Note that track.get_duration() returns a timedelta object, which
        # could (theoretically) also contain minutes, hours and days, however
        # the track.set_duration() method only accepts seconds, therefore it's
        # safe to read track.get_duration().seconds. Python 2.7 will bring the
        # proper solution with its timedelta.total_seconds() method.
        for k, scrobbler in self.scrobblers.items():
            scrobbler.report_now_playing(
                track.artist,
                track.title,
                track.album,
                str(track.get_duration().seconds),
                str(track.track),
            )

    def track_finished(self, track):
        logger.info("Track duration was: %s" % track.get_duration().seconds)

        # if track.get_duration() < datetime.timedelta(seconds=30):
        if track.get_duration().seconds < 30:
            # respecting last.fm minimum song lenght policy of 30 seconds
            logger.info("Track duration is < 30s, skip scrobble submit")
            return False

        if track.has_default_artist() or track.has_default_title():
            logger.error("Default artist or title, skip scrobble submit")
            return False

        logger.info("AS submit for track: %s - %s" % (track.artist, track.title))

        # Get UTC unix timestamp without daylight saving informations
        # http://stackoverflow.com/questions/1077285/how-to-specify-time-zone-utc-when-converting-to-unix-time-python
        timestamp = calendar.timegm(track.starttime.utctimetuple())

        # scrobbling to all networks
        #
        # Note that track.get_duration() returns a timedelta object, which
        # could (theoretically) also contain minutes, hours and days, however
        # the track.set_duration() method only accepts seconds, therefore it's
        # safe to read track.get_duration().seconds. Python 2.7 will bring the
        # proper solution with its timedelta.total_seconds() method.
        for k, scrobbler in self.scrobblers.items():
            scrobbler.scrobble(
                track.artist,
                track.title,
                int(timestamp),
                "P",
                "",
                str(track.get_duration().seconds),
                track.album,
                str(track.track),
            )


class IcecastTrackObserver(TrackObserver):
    """Update track metadata on an icecast mountpoint."""

    name = "Icecast"

    def __init__(self, baseUrl):
        self.baseUrl = baseUrl + "&mode=updinfo&charset=utf-8&song="

    def track_started(self, track):
        logger.info(
            "Updating Icecast Metadata for track: %s - %s" % (track.artist, track.title)
        )

        title = track.title

        if track.has_default_title() and track.has_default_artist():
            logger.info(
                "%s: Track has default info, using show instead" % self.__class__
            )

            title = track.show.name

        song_string = urllib.parse.quote_plus("%s - %s" % (track.artist, title))

        update_url = self.baseUrl + song_string

        logger.info("Icecast Update URL: " + update_url)

        urllib.request.urlopen(update_url)

    def track_finished(self, track):
        return True


class TickerTrackObserver(TrackObserver):
    """Writes the new ticker feed file.

    The feed file will be consumed by the song-ticker on the RaBe website. This is the
    successor of the long gone PseudoRssTrackObserver format used by the pre-WordPress
    website. This version here gets consumed by the WordPress website.
    """

    name = "Ticker"

    def __init__(self, ticker_file_path):
        warnings.warn(
            "The XML ticker format will be replaced with a JSON variant in the future",
            PendingDeprecationWarning,
        )
        self.ticker_file_path = ticker_file_path

    def track_started(self, track):
        logger.info(
            "Updating Ticker XML file for track: %s - %s" % (track.artist, track.title)
        )
        try:
            tz = pytz.timezone("Europe/Zurich")
        except pytz.exceptions.UnknownTimeZoneError:  # pragma: no coverage due to not knowing how to trigger
            tz = pytz.timezone("UTC")

        now = isodate.datetime_isoformat(datetime.datetime.now(tz))

        MAIN_NAMESPACE = "http://rabe.ch/schema/ticker.xsd"
        XLINK_NAMESPACE = "http://www.w3.org/1999/xlink"
        XLINK = "{%s}" % XLINK_NAMESPACE

        E = lxml.builder.ElementMaker(
            namespace=MAIN_NAMESPACE,
            nsmap={None: MAIN_NAMESPACE, "xlink": XLINK_NAMESPACE},
        )
        show_ref = E.link(track.show.url)
        show_ref.attrib[XLINK + "type"] = "simple"
        show_ref.attrib[XLINK + "href"] = track.show.url
        show_ref.attrib[XLINK + "show"] = "replace"

        ticker = E.ticker(
            E.identifier("ticker-%s" % uuid.uuid4()),
            E.creator("now-playing daemon v2"),
            E.date(now),
            E.show(
                E.name(track.show.name),
                show_ref,
                E.startTime(
                    isodate.datetime_isoformat(track.show.starttime.astimezone(tz))
                ),
                E.endTime(
                    isodate.datetime_isoformat(track.show.endtime.astimezone(tz))
                ),
                id=track.show.uuid,
            ),
            E.track(
                E.show(track.show.name, ref=track.show.uuid),
                E.artist(track.artist),
                E.title(track.title),
                E.startTime(isodate.datetime_isoformat(track.starttime.astimezone(tz))),
                E.endTime(isodate.datetime_isoformat(track.endtime.astimezone(tz))),
                id=track.uuid,
            ),
        )
        lxml.etree.ElementTree(ticker).write(
            self.ticker_file_path,
            pretty_print=True,
            xml_declaration=True,
            encoding="utf-8",
        )

    def track_finished(self, track):
        return True


class DabAudioCompanionTrackObserver(TrackObserver):
    """Update track metadata in a DAB+ tranmission through the 'Audio Companion' API."""

    name = "DAB+ Audio Companion"

    def __init__(self, baseUrl):
        self.baseUrl = baseUrl + "/api/setDLS?dls="

    def track_started(self, track):
        logger.info(
            "Updating DAB+ DLS for track: %s - %s" % (track.artist, track.title)
        )

        title = track.title

        if track.has_default_title() and track.has_default_artist():
            logger.info(
                "%s: Track has default info, using show instead" % self.__class__
            )

            title = track.show.name

        # artist is an unicode string which we have to encode into UTF-8
        # http://bugs.python.org/issue216716
        song_string = urllib.parse.quote_plus(
            "%s - %s" % (track.artist.encode("utf8"), title.encode("utf8"))
        )

        update_url = self.baseUrl + song_string

        logger.info("DAB+ Audio Companion URL: " + update_url)

        urllib.request.urlopen(update_url)

    def track_finished(self, track):
        return True



import logging
import logging.handlers
import select
import socket
import time
import warnings

logger = logging.getLogger(__name__)


class SaemuBoxError(Exception):
    """SaemuBox related exception."""

    pass


class SaemuBox:
    """Receive and validate info from Sämu Box for nowplaying."""

    output_mapping = {
        1: "Klangbecken",
        2: "Live + Replay",
        3: "Frei",
        4: "Vorproduktion",
        5: "Hörmal",
        6: "Studio Live",
    }

    def __init__(self):
        warnings.warn(
            "Saemubox will be replaced with Pathfinder", PendingDeprecationWarning
        )
        self.output = ""

        # listening ip adress (all)
        self.bind_ip = "0.0.0.0"

        # listening port
        self.port = 4001

        # allowed sender ip addresses
        self.senders = set(["10.130.225.6", "10.130.225.7"])

        # valid saemubox ids
        self.valid_ids = [str(i) for i in self.output_mapping]

        self._setup_socket()

        # wait for some data to arrive
        time.sleep(0.2)

    def _setup_socket(self):
        try:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.sock.bind((self.bind_ip, self.port))
            logger.info("SaemuBox: listening on %s:%i." % (self.bind_ip, self.port))
        except socket.error as e:
            self.sock = None
            logger.error("SaemuBox: cannot bind to %s:%i." % (self.bind_ip, self.port))
            raise SaemuBoxError(e.message)

    def __update(self):
        if self.sock is None or (hasattr(self.sock, "_closed") and self.sock._closed):
            logger.warn("SaemuBox: socket closed unexpectedly, retrying...")
            self._setup_socket()

        output = None
        seen_senders = set([])

        # read from socket while there is something to read (non-blocking)
        while select.select([self.sock], [], [], 0)[0]:
            data, addr = self.sock.recvfrom(1024)
            if addr[0] not in self.senders:
                logger.warn("SaemuBox: receiving data from invalid host: %s " % addr[0])
                continue

            ids = data.split()  # several saemubox ids might come in one packet
            if ids:
                id = ids[-1].decode("utf-8")  # only take last id
                if id in self.valid_ids:
                    seen_senders.add(addr[0])
                    output = id
                else:
                    logger.warn("SaemuBox: received invalid data: %s" % data)

        if output is None:
            logger.error("SaemuBox: could not read current status.")
            output = 0
            raise SaemuBoxError("Cannot read data from SaemuBox")
        elif seen_senders != self.senders:
            for missing_sender in self.senders - seen_senders:
                logger.warn("SaemuBox: missing sender: %s" % missing_sender)

        self.output = int(output)

    def get_active_output_id(self):
        self.__update()
        return self.output

    def get_active_output_name(self):
        self.__update()
        return self.output_mapping[self.output]

    def get_id_as_name(self, number):
        return self.output_mapping[number]


if __name__ == "__main__":
    # Test code
    import sys
    from random import randint

    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.setLevel(logging.INFO)

    sb = SaemuBox()
    localhost = socket.gethostbyname(socket.gethostname())
    sb.senders.add(localhost)

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    addr = (localhost, 4001)

    sock.sendto(b"\n", addr)
    sock.sendto(b"1", addr)
    sock.sendto(b"gugus", addr)  # invalid
    sock.sendto(b"1\n2\n1\n", addr)

    while True:
        sock.sendto(b"%d" % randint(1, 6), addr)
        sock.sendto(b"\n", addr)
        sock.sendto(b"\n", addr)
        sock.sendto(b" %d \n%d\n" % (randint(1, 6), randint(1, 6)), addr)
        sock.sendto(b"\n", addr)
        logger.info("Current status: %s" % sb.get_active_output_name())
        sock.sendto(b"\n", addr)
        sock.sendto(b"\n", addr)
        sock.sendto(b"%d" % randint(0, 7), addr)  # sometimes invalid
        time.sleep(1)



"""Track event handling subject of the observer."""

import logging
import logging.handlers

from track.observer import TrackObserver
from track.track import Track

logger = logging.getLogger(__name__)


class TrackEventHandler:
    """Inform all registered track-event observers about a track change.

    This is the subject of the classical observer pattern
    """

    def __init__(self):
        """Initialize the track event handler."""
        self.__observers = []

    def register_observer(self, observer: TrackObserver):
        """Register an observer to be informed about track changes."""
        logger.info("Registering TrackObserver '%s'", observer.__class__.__name__)
        self.__observers.append(observer)

    def remove_observer(self, observer: TrackObserver):
        """Remove an observer from the list of observers."""
        self.__observers.remove(observer)

    def get_observers(self) -> list:
        """Return register observers to allow inspecting them."""
        return self.__observers

    def track_started(self, track: Track):
        """Inform all registered track-event observers about a track started event."""
        logger.info("Track started: %s", track)

        for observer in self.__observers:
            logger.debug(
                "Sending track-started event to observer %s", observer.__class__
            )

            try:
                observer.track_started(track)
            except Exception as error:
                logger.exception(error)

    def track_finished(self, track: Track):
        """Inform all registered track-event observers about a track finished event."""
        logger.info("Track finished: %s", track)

        for observer in self.__observers:
            logger.info(
                "Sending track-finished event to observer %s", observer.__class__
            )

            try:
                observer.track_finished(track)
            except Exception as error:
                logger.exception(error)



import logging
import os
import signal
import sys
import time

from input import observer as inputObservers
from input.handler import InputHandler
from misc.saemubox import SaemuBox
from track import observer as trackObservers
from track.handler import TrackEventHandler

logger = logging.getLogger(__name__)


class NowPlayingDaemon:
    """The daemon part of the nowplaying app runs at all time and coordinates the input and output."""

    """initialize last_input to a know value."""
    last_input = 1

    def __init__(self, options):
        self.options = options

        self.saemubox = SaemuBox()

    def main(self):
        logger.info("Starting up now-playing daemon")

        try:
            self.register_signal_handlers()

            input_handler = self.get_input_handler()
        except Exception as e:
            logger.exception("Error: %s", e)
            sys.exit(-1)

        while True:
            try:
                saemubox_id = self.poll_saemubox()

                input_handler.update(saemubox_id)
            except Exception as e:
                logger.exception("Error: %s", e)

            time.sleep(self.options.sleepSeconds)

    def register_signal_handlers(self):
        logger.debug("Registering signal handler")
        signal.signal(signal.SIGINT, self.signal_handler)
        # signal.signal(signal.SIGKIL, self.signal_handler)

    def signal_handler(self, signum, frame):
        logger.debug("Signal %i caught" % signum)

        if signum == signal.SIGINT or signum == signal.SIGKILL:
            logger.info("Signal %i caught, terminating." % signum)
            sys.exit(os.EX_OK)

    def get_track_handler(self):
        handler = TrackEventHandler()
        [
            handler.register_observer(trackObservers.IcecastTrackObserver(url))
            for url in self.options.icecast
        ]
        [
            handler.register_observer(
                trackObservers.DabAudioCompanionTrackObserver(url)
            )
            for url in self.options.dab
        ]
        handler.register_observer(
            trackObservers.TickerTrackObserver(self.options.tickerOutputFile)
        )

        return handler

    def get_input_handler(self):
        handler = InputHandler()

        klangbecken = inputObservers.KlangbeckenInputObserver(
            self.options.currentShowUrl, self.options.inputFile
        )
        klangbecken.add_track_handler(self.get_track_handler())
        handler.register_observer(klangbecken)

        nonklangbecken = inputObservers.NonKlangbeckenInputObserver(
            self.options.currentShowUrl
        )
        nonklangbecken.add_track_handler(self.get_track_handler())
        handler.register_observer(nonklangbecken)

        return handler

    def poll_saemubox(self):
        """
        Poll Saemubox for new data.

        Should be run once per main loop.
        """

        saemubox_id = self.saemubox.get_active_output_id()
        logger.debug("Sämubox id: %i" % saemubox_id)

        if self.last_input != saemubox_id:
            logger.info(
                'Sämubox changed from "%s" to "%s"'
                % (
                    self.saemubox.get_id_as_name(self.last_input),
                    self.saemubox.get_id_as_name(saemubox_id),
                )
            )

        self.last_input = saemubox_id
        return saemubox_id



"""Set up nowplaying."""

from setuptools import find_packages, setup

with open("requirements.txt") as file:
    requirements = file.read().splitlines()


setup(
    name="rabe-nowplaying",
    description="Now Playing RaBe Songticker",
    url="http://github.com/radiorabe/nowplaying",
    author="RaBe IT-Reaktion",
    author_email="it@rabe.ch",
    license="AGPL-3",
    version_config=True,
    setup_requires=["setuptools-git-versioning"],
    install_requires=requirements,
    packages=find_packages(exclude=("tests",)),
    entry_points={"console_scripts": ["now-playing=nowplaying.cli:main"]},
    zip_safe=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU Affero General Public License v3",
    ],
)



import logging
import socket
import sys
import urllib.error
import urllib.parse
import urllib.request

from daemon import NowPlayingDaemon
from options import Options


class NowPlaying:
    def run(self):
        """Load configuration, initialize environment and start nowplaying daemon."""
        self.options = Options()
        self.options.parse_known_args()

        self.setup_logging()
        self.setup_urllib()
        socket.setdefaulttimeout(self.options.socketDefaultTimeout)

        NowPlayingDaemon(self.options).main()

    def setup_logging(self):
        """Configure logging to stdout."""
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

        stdout_handler = logging.StreamHandler(sys.stdout)

        stdout_formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )

        stdout_handler.setLevel(logging.INFO)
        if self.options.debug:
            stdout_handler.setLevel(logging.DEBUG)
        stdout_handler.setFormatter(stdout_formatter)

        root.addHandler(stdout_handler)

    def setup_urllib(self):
        """Prime urllib password manager for icecast auth."""
        http_password_manager = urllib.request.HTTPPasswordMgrWithDefaultRealm()

        http_password_manager.add_password(
            None, self.options.icecastBase, "source", self.options.icecastPassword
        )

        urllib.request.install_opener(
            urllib.request.build_opener(
                urllib.request.HTTPBasicAuthHandler(http_password_manager)
            )
        )


if __name__ == "__main__":
    NowPlaying().run()
